# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import datetime
from unittest import mock

from oslo_config import cfg
from oslo_utils import timeutils
from oslo_utils import uuidutils
from wsme import types as wtypes

from magnum.api import attr_validator
from magnum.api.controllers.v1 import bay as api_bay
from magnum.common import exception
from magnum.conductor import api as rpcapi
from magnum import objects
from magnum.tests import base
from magnum.tests.unit.api import base as api_base
from magnum.tests.unit.api import utils as apiutils
from magnum.tests.unit.db import utils as db_utils
from magnum.tests.unit.objects import utils as obj_utils


class TestBayObject(base.TestCase):

    def test_bay_init(self):
        bay_dict = apiutils.bay_post_data(baymodel_id=None)
        del bay_dict['node_count']
        del bay_dict['master_count']
        del bay_dict['bay_create_timeout']
        bay = api_bay.Bay(**bay_dict)
        self.assertEqual(1, bay.node_count)
        self.assertEqual(1, bay.master_count)
        self.assertEqual(60, bay.bay_create_timeout)

        # test unset value for baymodel_id
        bay.baymodel_id = wtypes.Unset
        self.assertEqual(wtypes.Unset, bay.baymodel_id)

        # test backwards compatibility of bay fields with new objects
        bay_dict['bay_create_timeout'] = 15
        bay_dict['bay_faults'] = {'testfault': 'fault'}
        bay = api_bay.Bay(**bay_dict)
        self.assertEqual(15, bay.bay_create_timeout)
        self.assertEqual(15, bay.create_timeout)
        self.assertIn('testfault', bay.bay_faults)
        self.assertIn('testfault', bay.faults)

    def test_as_dict_faults(self):
        bay_dict = apiutils.bay_post_data(baymodel_id=None)
        del bay_dict['node_count']
        del bay_dict['master_count']
        del bay_dict['bay_create_timeout']
        bay = api_bay.Bay(**bay_dict)
        bay.bay_faults = {'testfault': 'fault'}
        dict = bay.as_dict()
        self.assertEqual({'testfault': 'fault'}, dict['faults'])


class TestListBay(api_base.FunctionalTest):

    _bay_attrs = ("name", "baymodel_id", "node_count", "status",
                  "master_count", "stack_id", "bay_create_timeout")

    _expand_bay_attrs = ("name", "baymodel_id", "node_count", "status",
                         "api_address", "discovery_url", "node_addresses",
                         "master_count", "master_addresses", "stack_id",
                         "bay_create_timeout", "status_reason")

    def setUp(self):
        super(TestListBay, self).setUp()
        obj_utils.create_test_cluster_template(self.context)

    def test_empty(self):
        response = self.get_json('/bays')
        self.assertEqual([], response['bays'])

    def test_one(self):
        bay = obj_utils.create_test_cluster(self.context)
        response = self.get_json('/bays')
        self.assertEqual(bay.uuid, response['bays'][0]["uuid"])
        self._verify_attrs(self._bay_attrs, response['bays'][0])

        # Verify atts that should not appear from bay's get_all response
        none_attrs = set(self._expand_bay_attrs) - set(self._bay_attrs)
        self._verify_attrs(none_attrs, response['bays'][0], positive=False)

    def test_get_one(self):
        bay = obj_utils.create_test_cluster(self.context)
        response = self.get_json('/bays/%s' % bay['uuid'])
        self.assertEqual(bay.uuid, response['uuid'])
        self._verify_attrs(self._expand_bay_attrs, response)

    @mock.patch('magnum.common.clients.OpenStackClients.heat')
    def test_get_one_failed_bay(self, mock_heat):
        fake_resources = mock.MagicMock()
        fake_resources.resource_name = 'fake_name'
        fake_resources.resource_status_reason = 'fake_reason'

        ht = mock.MagicMock()
        ht.resources.list.return_value = [fake_resources]
        mock_heat.return_value = ht

        bay = obj_utils.create_test_cluster(self.context,
                                            status='CREATE_FAILED')
        response = self.get_json('/bays/%s' % bay['uuid'])
        self.assertEqual(bay.uuid, response['uuid'])
        self.assertEqual({'fake_name': 'fake_reason'}, response['bay_faults'])

    @mock.patch('magnum.common.clients.OpenStackClients.heat')
    def test_get_one_failed_bay_heatclient_exception(self, mock_heat):
        mock_heat.resources.list.side_effect = Exception('fake')
        bay = obj_utils.create_test_cluster(self.context,
                                            status='CREATE_FAILED')
        response = self.get_json('/bays/%s' % bay['uuid'])
        self.assertEqual(bay.uuid, response['uuid'])
        self.assertEqual({}, response['bay_faults'])

    def test_get_one_by_name(self):
        bay = obj_utils.create_test_cluster(self.context)
        response = self.get_json('/bays/%s' % bay['name'])
        self.assertEqual(bay.uuid, response['uuid'])
        self._verify_attrs(self._expand_bay_attrs, response)

    def test_get_one_by_name_not_found(self):
        response = self.get_json(
            '/bays/not_found',
            expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_get_one_by_name_multiple_bay(self):
        obj_utils.create_test_cluster(self.context, name='test_bay',
                                      uuid=uuidutils.generate_uuid())
        obj_utils.create_test_cluster(self.context, name='test_bay',
                                      uuid=uuidutils.generate_uuid())
        response = self.get_json('/bays/test_bay', expect_errors=True)
        self.assertEqual(409, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_get_all_with_pagination_marker(self):
        bay_list = []
        for id_ in range(4):
            bay = obj_utils.create_test_cluster(self.context, id=id_,
                                                uuid=uuidutils.generate_uuid())
            bay_list.append(bay)

        response = self.get_json('/bays?limit=3&marker=%s'
                                 % bay_list[2].uuid)
        self.assertEqual(1, len(response['bays']))
        self.assertEqual(bay_list[-1].uuid, response['bays'][0]['uuid'])

    def test_detail(self):
        bay = obj_utils.create_test_cluster(self.context)
        response = self.get_json('/bays/detail')
        self.assertEqual(bay.uuid, response['bays'][0]["uuid"])
        self._verify_attrs(self._expand_bay_attrs, response['bays'][0])

    def test_detail_with_pagination_marker(self):
        bay_list = []
        for id_ in range(4):
            bay = obj_utils.create_test_cluster(self.context, id=id_,
                                                uuid=uuidutils.generate_uuid())
            bay_list.append(bay)

        response = self.get_json('/bays/detail?limit=3&marker=%s'
                                 % bay_list[2].uuid)
        self.assertEqual(1, len(response['bays']))
        self.assertEqual(bay_list[-1].uuid, response['bays'][0]['uuid'])
        self._verify_attrs(self._expand_bay_attrs, response['bays'][0])

    def test_detail_against_single(self):
        bay = obj_utils.create_test_cluster(self.context)
        response = self.get_json('/bays/%s/detail' % bay['uuid'],
                                 expect_errors=True)
        self.assertEqual(404, response.status_int)

    def test_many(self):
        bm_list = []
        for id_ in range(5):
            bay = obj_utils.create_test_cluster(self.context, id=id_,
                                                uuid=uuidutils.generate_uuid())
            bm_list.append(bay.uuid)
        response = self.get_json('/bays')
        self.assertEqual(len(bm_list), len(response['bays']))
        uuids = [b['uuid'] for b in response['bays']]
        self.assertEqual(sorted(bm_list), sorted(uuids))

    def test_links(self):
        uuid = uuidutils.generate_uuid()
        obj_utils.create_test_cluster(self.context, id=1, uuid=uuid)
        response = self.get_json('/bays/%s' % uuid)
        self.assertIn('links', response.keys())
        self.assertEqual(2, len(response['links']))
        self.assertIn(uuid, response['links'][0]['href'])
        for link in response['links']:
            bookmark = link['rel'] == 'bookmark'
            self.assertTrue(self.validate_link(link['href'],
                                               bookmark=bookmark))

    def test_collection_links(self):
        for id_ in range(5):
            obj_utils.create_test_cluster(self.context, id=id_,
                                          uuid=uuidutils.generate_uuid())
        response = self.get_json('/bays/?limit=3')
        self.assertEqual(3, len(response['bays']))

        next_marker = response['bays'][-1]['uuid']
        self.assertIn(next_marker, response['next'])

    def test_collection_links_default_limit(self):
        cfg.CONF.set_override('max_limit', 3, 'api')
        for id_ in range(5):
            obj_utils.create_test_cluster(self.context, id=id_,
                                          uuid=uuidutils.generate_uuid())
        response = self.get_json('/bays')
        self.assertEqual(3, len(response['bays']))

        next_marker = response['bays'][-1]['uuid']
        self.assertIn(next_marker, response['next'])


class TestPatch(api_base.FunctionalTest):

    def setUp(self):
        super(TestPatch, self).setUp()
        self.cluster_template = obj_utils.create_test_cluster_template(
            self.context)
        self.bay = obj_utils.create_test_cluster(self.context,
                                                 name='bay_example_A',
                                                 node_count=3)
        p = mock.patch.object(rpcapi.API, 'cluster_update')
        self.mock_bay_update = p.start()
        self.mock_bay_update.side_effect = self._simulate_rpc_bay_update
        self.addCleanup(p.stop)

    def _simulate_rpc_bay_update(self, bay, node_count, rollback=False):
        bay.status = 'UPDATE_IN_PROGRESS'
        bay.save()
        default_ng_worker = bay.default_ng_worker
        default_ng_worker.node_count = node_count
        default_ng_worker.save()
        return bay

    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_replace_ok(self, mock_utcnow):
        new_node_count = 4
        test_time = datetime.datetime(2000, 1, 1, 0, 0)
        mock_utcnow.return_value = test_time

        response = self.patch_json('/bays/%s' % self.bay.uuid,
                                   [{'path': '/node_count',
                                     'value': new_node_count,
                                     'op': 'replace'}])
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(200, response.status_code)

        response = self.get_json('/bays/%s' % self.bay.uuid)
        self.assertEqual(new_node_count, response['node_count'])
        return_updated_at = timeutils.parse_isotime(
            response['updated_at']).replace(tzinfo=None)
        self.assertEqual(test_time, return_updated_at)
        # Assert nothing else was changed
        self.assertEqual(self.bay.uuid, response['uuid'])
        self.assertEqual(self.bay.cluster_template_id, response['baymodel_id'])

    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_replace_ok_by_name(self, mock_utcnow):
        new_node_count = 4
        test_time = datetime.datetime(2000, 1, 1, 0, 0)
        mock_utcnow.return_value = test_time

        response = self.patch_json('/bays/%s' % self.bay.name,
                                   [{'path': '/node_count',
                                     'value': new_node_count,
                                     'op': 'replace'}])
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(200, response.status_code)

        response = self.get_json('/bays/%s' % self.bay.uuid)
        self.assertEqual(new_node_count, response['node_count'])
        return_updated_at = timeutils.parse_isotime(
            response['updated_at']).replace(tzinfo=None)
        self.assertEqual(test_time, return_updated_at)
        # Assert nothing else was changed
        self.assertEqual(self.bay.uuid, response['uuid'])
        self.assertEqual(self.bay.cluster_template_id, response['baymodel_id'])

    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_replace_ok_by_name_not_found(self, mock_utcnow):
        name = 'not_found'
        test_time = datetime.datetime(2000, 1, 1, 0, 0)
        mock_utcnow.return_value = test_time

        response = self.patch_json('/bays/%s' % name,
                                   [{'path': '/name', 'value': name,
                                     'op': 'replace'}],
                                   expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(404, response.status_code)

    def test_replace_baymodel_id_failed(self):
        cluster_template = obj_utils.create_test_cluster_template(
            self.context,
            uuid=uuidutils.generate_uuid())
        response = self.patch_json('/bays/%s' % self.bay.uuid,
                                   [{'path': '/baymodel_id',
                                     'value': cluster_template.uuid,
                                     'op': 'replace'}],
                                   expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_code)
        self.assertTrue(response.json['errors'])

    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_replace_ok_by_name_multiple_bay(self, mock_utcnow):
        test_time = datetime.datetime(2000, 1, 1, 0, 0)
        mock_utcnow.return_value = test_time

        obj_utils.create_test_cluster(self.context, name='test_bay',
                                      uuid=uuidutils.generate_uuid())
        obj_utils.create_test_cluster(self.context, name='test_bay',
                                      uuid=uuidutils.generate_uuid())

        response = self.patch_json('/bays/test_bay',
                                   [{'path': '/name', 'value': 'test_bay',
                                     'op': 'replace'}],
                                   expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(409, response.status_code)

    def test_replace_non_existent_baymodel_id(self):
        response = self.patch_json('/bays/%s' % self.bay.uuid,
                                   [{'path': '/baymodel_id',
                                     'value': uuidutils.generate_uuid(),
                                     'op': 'replace'}],
                                   expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_code)
        self.assertTrue(response.json['errors'])

    def test_replace_invalid_node_count(self):
        response = self.patch_json('/bays/%s' % self.bay.uuid,
                                   [{'path': '/node_count', 'value': -1,
                                     'op': 'replace'}],
                                   expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_code)
        self.assertTrue(response.json['errors'])

    def test_replace_non_existent_bay(self):
        response = self.patch_json('/bays/%s' % uuidutils.generate_uuid(),
                                   [{'path': '/name',
                                     'value': 'bay_example_B',
                                     'op': 'replace'}],
                                   expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_replace_bay_name_failed(self):
        response = self.patch_json('/bays/%s' % self.bay.uuid,
                                   [{'path': '/name',
                                     'value': 'bay_example_B',
                                     'op': 'replace'}],
                                   expect_errors=True)
        self.assertEqual(400, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_add_non_existent_property(self):
        response = self.patch_json(
            '/bays/%s' % self.bay.uuid,
            [{'path': '/foo', 'value': 'bar', 'op': 'add'}],
            expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    @mock.patch.object(rpcapi.API, 'cluster_update_async')
    def test_update_bay_async(self, mock_update):
        response = self.patch_json(
            '/bays/%s' % self.bay.name,
            [{'path': '/node_count', 'value': 4,
              'op': 'replace'}],
            headers={'OpenStack-API-Version': 'container-infra 1.2'})

        self.assertEqual(202, response.status_code)

    @mock.patch.object(rpcapi.API, 'cluster_update_async')
    def test_update_bay_with_rollback_enabled(self, mock_update):
        node_count = 4
        response = self.patch_json(
            '/bays/%s/?rollback=True' % self.bay.name,
            [{'path': '/node_count', 'value': node_count,
              'op': 'replace'}],
            headers={'OpenStack-API-Version': 'container-infra 1.3'})

        mock_update.assert_called_once_with(mock.ANY, node_count,
                                            rollback=True)
        self.assertEqual(202, response.status_code)

    def test_remove_ok(self):
        response = self.get_json('/bays/%s' % self.bay.uuid)
        self.assertIsNotNone(response['name'])

        response = self.patch_json('/bays/%s' % self.bay.uuid,
                                   [{'path': '/node_count', 'op': 'remove'}])
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(200, response.status_code)

        response = self.get_json('/bays/%s' % self.bay.uuid)
        # only allow node_count for bay, and default value is 1
        self.assertEqual(1, response['node_count'])
        # Assert nothing else was changed
        self.assertEqual(self.bay.uuid, response['uuid'])
        self.assertEqual(self.bay.cluster_template_id, response['baymodel_id'])
        self.assertEqual(self.bay.name, response['name'])
        self.assertEqual(self.bay.master_count, response['master_count'])

    def test_remove_mandatory_property_fail(self):
        mandatory_properties = ('/uuid', '/baymodel_id')
        for p in mandatory_properties:
            response = self.patch_json('/bays/%s' % self.bay.uuid,
                                       [{'path': p, 'op': 'remove'}],
                                       expect_errors=True)
            self.assertEqual(400, response.status_int)
            self.assertEqual('application/json', response.content_type)
            self.assertTrue(response.json['errors'])

    def test_remove_non_existent_property(self):
        response = self.patch_json(
            '/bays/%s' % self.bay.uuid,
            [{'path': '/non-existent', 'op': 'remove'}],
            expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_code)
        self.assertTrue(response.json['errors'])


class TestPost(api_base.FunctionalTest):

    def setUp(self):
        super(TestPost, self).setUp()
        self.cluster_template = obj_utils.create_test_cluster_template(
            self.context)
        p = mock.patch.object(rpcapi.API, 'cluster_create')
        self.mock_bay_create = p.start()
        self.mock_bay_create.side_effect = self._simulate_rpc_bay_create
        self.addCleanup(p.stop)
        p = mock.patch.object(attr_validator, 'validate_os_resources')
        self.mock_valid_os_res = p.start()
        self.addCleanup(p.stop)

    def _simulate_rpc_bay_create(self, bay, master_count, node_count,
                                 bay_create_timeout):
        bay.create()
        db_utils.create_nodegroups_for_cluster(
            cluster_id=bay.uuid, node_count=node_count,
            master_count=master_count)
        return bay

    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_create_bay(self, mock_utcnow):
        bdict = apiutils.bay_post_data()
        test_time = datetime.datetime(2000, 1, 1, 0, 0)
        mock_utcnow.return_value = test_time

        response = self.post_json('/bays', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)
        # Check location header
        self.assertIsNotNone(response.location)
        self.assertTrue(uuidutils.is_uuid_like(response.json['uuid']))
        self.assertNotIn('updated_at', response.json.keys)
        return_created_at = timeutils.parse_isotime(
            response.json['created_at']).replace(tzinfo=None)
        self.assertEqual(test_time, return_created_at)
        self.assertEqual(bdict['bay_create_timeout'],
                         response.json['bay_create_timeout'])

    def test_create_bay_set_project_id_and_user_id(self):
        bdict = apiutils.bay_post_data()

        def _simulate_rpc_bay_create(bay, node_count, master_count,
                                     bay_create_timeout):
            self.assertEqual(self.context.project_id, bay.project_id)
            self.assertEqual(self.context.user_id, bay.user_id)
            bay.create()
            db_utils.create_nodegroups_for_cluster(
                cluster_id=bay.uuid, node_count=node_count,
                master_count=master_count)
            return bay
        self.mock_bay_create.side_effect = _simulate_rpc_bay_create

        self.post_json('/bays', bdict)

    def test_create_bay_doesnt_contain_id(self):
        with mock.patch.object(self.dbapi, 'create_cluster',
                               wraps=self.dbapi.create_cluster) as cc_mock:
            bdict = apiutils.bay_post_data(name='bay_example_A')
            response = self.post_json('/bays', bdict)
            self.assertEqual(bdict['name'], response.json['name'])
            cc_mock.assert_called_once_with(mock.ANY)
            # Check that 'id' is not in first arg of positional args
            self.assertNotIn('id', cc_mock.call_args[0][0])

    def test_create_bay_generate_uuid(self):
        bdict = apiutils.bay_post_data()
        del bdict['uuid']

        response = self.post_json('/bays', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)
        self.assertEqual(bdict['name'], response.json['name'])
        self.assertTrue(uuidutils.is_uuid_like(response.json['uuid']))

    def test_create_bay_no_baymodel_id(self):
        bdict = apiutils.bay_post_data()
        del bdict['baymodel_id']
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)

    def test_create_bay_with_non_existent_baymodel_id(self):
        bdict = apiutils.bay_post_data(baymodel_id=uuidutils.generate_uuid())
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_bay_with_baymodel_name(self):
        bdict = apiutils.bay_post_data(baymodel_id=self.cluster_template.name)
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)

    def test_create_bay_with_node_count_zero(self):
        bdict = apiutils.bay_post_data()
        bdict['node_count'] = 0
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_bay_with_node_count_negative(self):
        bdict = apiutils.bay_post_data()
        bdict['node_count'] = -1
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_bay_with_no_node_count(self):
        bdict = apiutils.bay_post_data()
        del bdict['node_count']
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)
        self.assertEqual(1, response.json['node_count'])

    def test_create_bay_with_master_count_zero(self):
        bdict = apiutils.bay_post_data()
        bdict['master_count'] = 0
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_bay_with_no_master_count(self):
        bdict = apiutils.bay_post_data()
        del bdict['master_count']
        response = self.post_json('/bays', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)
        self.assertEqual(1, response.json['master_count'])

    def test_create_bay_with_invalid_long_name(self):
        bdict = apiutils.bay_post_data(name='x' * 243)
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_bay_with_invalid_integer_name(self):
        bdict = apiutils.bay_post_data(name='123456')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_bay_with_invalid_integer_str_name(self):
        bdict = apiutils.bay_post_data(name='123456test_bay')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_bay_with_hyphen_invalid_at_start_name(self):
        bdict = apiutils.bay_post_data(name='-test_bay')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_bay_with_period_invalid_at_start_name(self):
        bdict = apiutils.bay_post_data(name='.test_bay')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_bay_with_underscore_invalid_at_start_name(self):
        bdict = apiutils.bay_post_data(name='_test_bay')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_bay_with_valid_str_int_name(self):
        bdict = apiutils.bay_post_data(name='test_bay123456')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)
        self.assertEqual(response.json['name'], bdict['name'])

    def test_create_bay_with_hyphen_valid_name(self):
        bdict = apiutils.bay_post_data(name='test-bay')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)
        self.assertEqual(response.json['name'], bdict['name'])

    def test_create_bay_with_period_valid_name(self):
        bdict = apiutils.bay_post_data(name='test.bay')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)
        self.assertEqual(response.json['name'], bdict['name'])

    def test_create_bay_with_period_at_end_valid_name(self):
        bdict = apiutils.bay_post_data(name='testbay.')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)
        self.assertEqual(response.json['name'], bdict['name'])

    def test_create_bay_with_hyphen_at_end_valid_name(self):
        bdict = apiutils.bay_post_data(name='testbay-')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)
        self.assertEqual(response.json['name'], bdict['name'])

    def test_create_bay_with_underscore_at_end_valid_name(self):
        bdict = apiutils.bay_post_data(name='testbay_')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)
        self.assertEqual(response.json['name'], bdict['name'])

    def test_create_bay_with_mix_special_char_valid_name(self):
        bdict = apiutils.bay_post_data(name='test.-_bay')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)
        self.assertEqual(response.json['name'], bdict['name'])

    def test_create_bay_with_capital_letter_start_valid_name(self):
        bdict = apiutils.bay_post_data(name='Testbay')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)
        self.assertEqual(response.json['name'], bdict['name'])

    def test_create_bay_with_invalid_empty_name(self):
        bdict = apiutils.bay_post_data(name='')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_bay_without_name(self):
        bdict = apiutils.bay_post_data()
        del bdict['name']
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)
        self.assertIsNotNone(response.json['name'])

    def test_create_bay_with_timeout_none(self):
        bdict = apiutils.bay_post_data()
        bdict['bay_create_timeout'] = None
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)

    def test_create_bay_with_no_timeout(self):
        def _simulate_rpc_bay_create(bay, node_count, master_count,
                                     bay_create_timeout):
            self.assertEqual(60, bay_create_timeout)
            bay.create()
            db_utils.create_nodegroups_for_cluster(
                cluster_id=bay.uuid, node_count=node_count,
                master_count=master_count)
            return bay
        self.mock_bay_create.side_effect = _simulate_rpc_bay_create
        bdict = apiutils.bay_post_data()
        del bdict['bay_create_timeout']
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)

    def test_create_bay_with_timeout_negative(self):
        bdict = apiutils.bay_post_data()
        bdict['bay_create_timeout'] = -1
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)
        self.assertTrue(response.json['errors'])

    def test_create_bay_with_timeout_zero(self):
        bdict = apiutils.bay_post_data()
        bdict['bay_create_timeout'] = 0
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)

    def test_create_bay_with_invalid_flavor(self):
        bdict = apiutils.bay_post_data()
        self.mock_valid_os_res.side_effect = exception.FlavorNotFound(
            'test-flavor')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(self.mock_valid_os_res.called)
        self.assertEqual(400, response.status_int)

    def test_create_bay_with_invalid_ext_network(self):
        bdict = apiutils.bay_post_data()
        self.mock_valid_os_res.side_effect = exception.ExternalNetworkNotFound(
            'test-net')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(self.mock_valid_os_res.called)
        self.assertEqual(400, response.status_int)

    def test_create_bay_with_invalid_keypair(self):
        bdict = apiutils.bay_post_data()
        self.mock_valid_os_res.side_effect = exception.KeyPairNotFound(
            'test-key')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(self.mock_valid_os_res.called)
        self.assertEqual(404, response.status_int)

    def test_create_bay_with_nonexist_image(self):
        bdict = apiutils.bay_post_data()
        self.mock_valid_os_res.side_effect = exception.ImageNotFound(
            'test-img')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(self.mock_valid_os_res.called)
        self.assertEqual(400, response.status_int)

    def test_create_bay_with_multi_images_same_name(self):
        bdict = apiutils.bay_post_data()
        self.mock_valid_os_res.side_effect = exception.Conflict('test-img')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(self.mock_valid_os_res.called)
        self.assertEqual(409, response.status_int)

    def test_create_bay_with_on_os_distro_image(self):
        bdict = apiutils.bay_post_data()
        self.mock_valid_os_res.side_effect = exception.OSDistroFieldNotFound(
            'img')
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(self.mock_valid_os_res.called)
        self.assertEqual(400, response.status_int)

    def test_create_bay_with_no_lb_one_node(self):
        cluster_template = obj_utils.create_test_cluster_template(
            self.context, name='foo', uuid='foo', master_lb_enabled=False)
        bdict = apiutils.bay_post_data(baymodel_id=cluster_template.name,
                                       master_count=1)
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)

    def test_create_bay_with_no_lb_multi_node(self):
        cluster_template = obj_utils.create_test_cluster_template(
            self.context, name='foo', uuid='foo', master_lb_enabled=False)
        bdict = apiutils.bay_post_data(baymodel_id=cluster_template.name,
                                       master_count=3, master_lb_enabled=False)
        response = self.post_json('/bays', bdict, expect_errors=True)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(400, response.status_int)

    def test_create_bay_with_docker_volume_size(self):
        bdict = apiutils.bay_post_data()
        bdict['docker_volume_size'] = 3
        response = self.post_json('/bays', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)
        bay, timeout = self.mock_bay_create.call_args
        self.assertEqual(3, bay[0].docker_volume_size)

    def test_create_bay_without_docker_volume_size(self):
        bdict = apiutils.bay_post_data()
        # Remove the default docker_volume_size from the bay dict.
        del bdict['docker_volume_size']
        response = self.post_json('/bays', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(201, response.status_int)
        bay, timeout = self.mock_bay_create.call_args
        # Verify docker_volume_size from BayModel is used
        self.assertEqual(20, bay[0].docker_volume_size)


class TestDelete(api_base.FunctionalTest):

    def setUp(self):
        super(TestDelete, self).setUp()
        self.cluster_template = obj_utils.create_test_cluster_template(
            self.context)
        self.bay = obj_utils.create_test_cluster(self.context)
        p = mock.patch.object(rpcapi.API, 'cluster_delete')
        self.mock_bay_delete = p.start()
        self.mock_bay_delete.side_effect = self._simulate_rpc_bay_delete
        self.addCleanup(p.stop)

    def _simulate_rpc_bay_delete(self, bay_uuid):
        bay = objects.Cluster.get_by_uuid(self.context, bay_uuid)
        bay.destroy()
        ngs = objects.NodeGroup.list(self.context, bay_uuid)
        for ng in ngs:
            ng.destroy()

    def test_delete_bay(self):
        self.delete('/bays/%s' % self.bay.uuid)
        response = self.get_json('/bays/%s' % self.bay.uuid,
                                 expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_delete_bay_not_found(self):
        uuid = uuidutils.generate_uuid()
        response = self.delete('/bays/%s' % uuid, expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_delete_bay_with_name_not_found(self):
        response = self.delete('/bays/not_found', expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_delete_bay_with_name(self):
        response = self.delete('/bays/%s' % self.bay.name,
                               expect_errors=True)
        self.assertEqual(204, response.status_int)

    def test_delete_multiple_bay_by_name(self):
        obj_utils.create_test_cluster(self.context, name='test_bay',
                                      uuid=uuidutils.generate_uuid())
        obj_utils.create_test_cluster(self.context, name='test_bay',
                                      uuid=uuidutils.generate_uuid())
        response = self.delete('/bays/test_bay', expect_errors=True)
        self.assertEqual(409, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])


class TestBayPolicyEnforcement(api_base.FunctionalTest):

    def setUp(self):
        super(TestBayPolicyEnforcement, self).setUp()
        obj_utils.create_test_cluster_template(self.context)

    def _common_policy_check(self, rule, func, *arg, **kwarg):
        self.policy.set_rules({rule: "project:non_fake"})
        response = func(*arg, **kwarg)
        self.assertEqual(403, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(
            "Policy doesn't allow %s to be performed." % rule,
            response.json['errors'][0]['detail'])

    def test_policy_disallow_get_all(self):
        self._common_policy_check(
            "bay:get_all", self.get_json, '/bays', expect_errors=True)

    def test_policy_disallow_get_one(self):
        self.bay = obj_utils.create_test_cluster(self.context)
        self._common_policy_check(
            "bay:get", self.get_json, '/bays/%s' % self.bay.uuid,
            expect_errors=True)

    def test_policy_disallow_detail(self):
        self._common_policy_check(
            "bay:detail", self.get_json,
            '/bays/%s/detail' % uuidutils.generate_uuid(),
            expect_errors=True)

    def test_policy_disallow_update(self):
        self.bay = obj_utils.create_test_cluster(self.context,
                                                 name='bay_example_A',
                                                 node_count=3)
        self._common_policy_check(
            "bay:update", self.patch_json, '/bays/%s' % self.bay.name,
            [{'path': '/name', 'value': "new_name", 'op': 'replace'}],
            expect_errors=True)

    def test_policy_disallow_create(self):
        bdict = apiutils.bay_post_data(name='bay_example_A')
        self._common_policy_check(
            "bay:create", self.post_json, '/bays', bdict, expect_errors=True)

    def _simulate_rpc_bay_delete(self, bay_uuid):
        bay = objects.Cluster.get_by_uuid(self.context, bay_uuid)
        bay.destroy()

    def test_policy_disallow_delete(self):
        p = mock.patch.object(rpcapi.API, 'cluster_delete')
        self.mock_bay_delete = p.start()
        self.mock_bay_delete.side_effect = self._simulate_rpc_bay_delete
        self.addCleanup(p.stop)
        self.bay = obj_utils.create_test_cluster(self.context)
        self._common_policy_check(
            "bay:delete", self.delete, '/bays/%s' % self.bay.uuid,
            expect_errors=True)

    def _owner_check(self, rule, func, *args, **kwargs):
        self.policy.set_rules({rule: "user_id:%(user_id)s"})
        response = func(*args, **kwargs)
        self.assertEqual(403, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(
            "Policy doesn't allow %s to be performed." % rule,
            response.json['errors'][0]['detail'])

    def test_policy_only_owner_get_one(self):
        bay = obj_utils.create_test_cluster(self.context, user_id='another')
        self._owner_check("bay:get", self.get_json, '/bays/%s' % bay.uuid,
                          expect_errors=True)

    def test_policy_only_owner_update(self):
        bay = obj_utils.create_test_cluster(self.context, user_id='another')
        self._owner_check(
            "bay:update", self.patch_json, '/bays/%s' % bay.uuid,
            [{'path': '/name', 'value': "new_name", 'op': 'replace'}],
            expect_errors=True)

    def test_policy_only_owner_delete(self):
        bay = obj_utils.create_test_cluster(self.context, user_id='another')
        self._owner_check("bay:delete", self.delete, '/bays/%s' % bay.uuid,
                          expect_errors=True)



# Copyright 2015 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import timeutils
import pecan
import wsme
from wsme import types as wtypes

from magnum.api.controllers import base
from magnum.api.controllers import link
from magnum.api.controllers.v1 import types
from magnum.api import expose
from magnum.api import utils as api_utils
from magnum.common import exception
from magnum.common import policy
from magnum import objects


class ClusterID(wtypes.Base):
    """API representation of a cluster ID

    This class enforces type checking and value constraints, and converts
    between the internal object model and the API representation of a cluster
    ID.
    """

    uuid = types.uuid
    """Unique UUID for this cluster"""

    def __init__(self, uuid):
        self.uuid = uuid


class Certificate(base.APIBase):
    """API representation of a certificate.

    This class enforces type checking and value constraints, and converts
    between the internal object model and the API representation of a
    certificate.
    """

    _cluster_uuid = None
    """uuid or logical name of cluster"""

    _cluster = None

    def _get_cluster_uuid(self):
        return self._cluster_uuid

    def _set_cluster_uuid(self, value):
        if value and self._cluster_uuid != value:
            try:
                self._cluster = api_utils.get_resource('Cluster', value)
                self._cluster_uuid = self._cluster.uuid
            except exception.ClusterNotFound as e:
                # Change error code because 404 (NotFound) is inappropriate
                # response for a POST request to create a Cluster
                e.code = 400  # BadRequest
                raise
        elif value == wtypes.Unset:
            self._cluster_uuid = wtypes.Unset

    bay_uuid = wsme.wsproperty(wtypes.text, _get_cluster_uuid,
                               _set_cluster_uuid)
    """The bay UUID or id"""

    cluster_uuid = wsme.wsproperty(wtypes.text, _get_cluster_uuid,
                                   _set_cluster_uuid)
    """The cluster UUID or id"""

    links = wsme.wsattr([link.Link], readonly=True)
    """A list containing a self link and associated certificate links"""

    csr = wtypes.StringType(min_length=1)
    """"The Certificate Signing Request"""

    pem = wtypes.StringType()
    """"The Signed Certificate"""

    def __init__(self, **kwargs):
        super(Certificate, self).__init__()

        self.fields = []
        for field in objects.Certificate.fields:
            # Skip fields we do not expose.
            if not hasattr(self, field):
                continue
            self.fields.append(field)
            setattr(self, field, kwargs.get(field, wtypes.Unset))

        # set the attribute for bay_uuid for backwards compatibility
        self.fields.append('bay_uuid')
        setattr(self, 'bay_uuid', kwargs.get('bay_uuid',  self._cluster_uuid))

    def get_cluster(self):
        if not self._cluster:
            self._cluster = api_utils.get_resource('Cluster',
                                                   self.cluster_uuid)
        return self._cluster

    @staticmethod
    def _convert_with_links(certificate, url, expand=True):
        if not expand:
            certificate.unset_fields_except(['bay_uuid', 'cluster_uuid',
                                             'csr', 'pem'])

        certificate.links = [link.Link.make_link('self', url,
                                                 'certificates',
                                                 certificate.cluster_uuid),
                             link.Link.make_link('bookmark', url,
                                                 'certificates',
                                                 certificate.cluster_uuid,
                                                 bookmark=True)]
        return certificate

    @classmethod
    def convert_with_links(cls, rpc_cert, expand=True):
        cert = Certificate(**rpc_cert.as_dict())
        return cls._convert_with_links(cert,
                                       pecan.request.host_url, expand)

    @classmethod
    def sample(cls, expand=True):
        sample = cls(bay_uuid='7ae81bb3-dec3-4289-8d6c-da80bd8001ae',
                     cluster_uuid='7ae81bb3-dec3-4289-8d6c-da80bd8001ae',
                     created_at=timeutils.utcnow(),
                     csr='AAA....AAA')
        return cls._convert_with_links(sample, 'http://localhost:9511', expand)


class CertificateController(base.Controller):
    """REST controller for Certificate."""

    def __init__(self):
        super(CertificateController, self).__init__()

    _custom_actions = {
        'detail': ['GET'],
    }

    @expose.expose(Certificate, types.uuid_or_name)
    def get_one(self, cluster_ident):
        """Retrieve CA information about the given cluster.

        :param cluster_ident: UUID of a cluster or
        logical name of the cluster.
        """
        context = pecan.request.context
        cluster = api_utils.get_resource('Cluster', cluster_ident)
        policy.enforce(context, 'certificate:get', cluster.as_dict(),
                       action='certificate:get')
        certificate = pecan.request.rpcapi.get_ca_certificate(cluster)
        return Certificate.convert_with_links(certificate)

    @expose.expose(Certificate, body=Certificate, status_code=201)
    def post(self, certificate):
        """Sign a new certificate by the CA.

        :param certificate: a certificate within the request body.
        """
        context = pecan.request.context
        cluster = certificate.get_cluster()
        policy.enforce(context, 'certificate:create', cluster.as_dict(),
                       action='certificate:create')
        certificate_dict = certificate.as_dict()
        certificate_dict['project_id'] = context.project_id
        certificate_dict['user_id'] = context.user_id
        cert_obj = objects.Certificate(context, **certificate_dict)

        new_cert = pecan.request.rpcapi.sign_certificate(cluster,
                                                         cert_obj)
        return Certificate.convert_with_links(new_cert)

    @expose.expose(ClusterID, types.uuid_or_name, status_code=202)
    def patch(self, cluster_ident):
        context = pecan.request.context
        cluster = api_utils.get_resource('Cluster', cluster_ident)
        policy.enforce(context, 'certificate:rotate_ca', cluster.as_dict(),
                       action='certificate:rotate_ca')
        if cluster.cluster_template.tls_disabled:
            raise exception.NotSupported("Rotating the CA certificate on a "
                                         "non-TLS cluster is not supported")

        pecan.request.rpcapi.rotate_ca_certificate(cluster)

        return ClusterID(cluster.uuid)



# Copyright 2015 NEC Corporation.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import oslo_messaging as messaging
from oslo_versionedobjects import fields

from magnum.conductor.handlers import indirection_api
from magnum.objects import base as obj_base
from magnum.tests import base


class TestIndirectionApiConductor(base.TestCase):
    def setUp(self):
        super(TestIndirectionApiConductor, self).setUp()
        self.conductor = indirection_api.Handler()

    def _test_object_action(self, is_classmethod, raise_exception):
        @obj_base.MagnumObjectRegistry.register
        class TestObject(obj_base.MagnumObject):
            def foo(self, context, raise_exception=False):
                if raise_exception:
                    raise Exception('test')
                else:
                    return 'test'

            @classmethod
            def bar(cls, context, raise_exception=False):
                if raise_exception:
                    raise Exception('test')
                else:
                    return 'test'

        obj = TestObject()
        if is_classmethod:
            result = self.conductor.object_class_action(
                self.context, TestObject.obj_name(), 'bar', '1.0',
                tuple(), {'raise_exception': raise_exception})
        else:
            updates, result = self.conductor.object_action(
                self.context, obj, 'foo', tuple(),
                {'raise_exception': raise_exception})
        self.assertEqual('test', result)

    def test_object_action(self):
        self._test_object_action(False, False)

    def test_object_action_on_raise(self):
        self.assertRaises(messaging.ExpectedException,
                          self._test_object_action, False, True)

    def test_object_class_action(self):
        self._test_object_action(True, False)

    def test_object_class_action_on_raise(self):
        self.assertRaises(messaging.ExpectedException,
                          self._test_object_action, True, True)

    def test_object_action_copies_object(self):
        @obj_base.MagnumObjectRegistry.register
        class TestObject(obj_base.MagnumObject):
            fields = {'dict': fields.DictOfStringsField()}

            def touch_dict(self, context):
                self.dict['foo'] = 'bar'
                self.obj_reset_changes()

        obj = TestObject()
        obj.dict = {}
        obj.obj_reset_changes()
        updates, result = self.conductor.object_action(
            self.context, obj, 'touch_dict', tuple(), {})
        # NOTE(danms): If conductor did not properly copy the object, then
        # the new and reference copies of the nested dict object will be
        # the same, and thus 'dict' will not be reported as changed
        self.assertIn('dict', updates)
        self.assertEqual({'foo': 'bar'}, updates['dict'])



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_policy import policy

from magnum.common.policies import base

CLUSTER_TEMPLATE = 'clustertemplate:%s'

rules = [
    policy.DocumentedRuleDefault(
        name=CLUSTER_TEMPLATE % 'create',
        check_str=base.RULE_DENY_CLUSTER_USER,
        description='Create a new cluster template.',
        operations=[
            {
                'path': '/v1/clustertemplates',
                'method': 'POST'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=CLUSTER_TEMPLATE % 'delete',
        check_str=base.RULE_DENY_CLUSTER_USER,
        description='Delete a cluster template.',
        operations=[
            {
                'path': '/v1/clustertemplate/{clustertemplate_ident}',
                'method': 'DELETE'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=CLUSTER_TEMPLATE % 'delete_all_projects',
        check_str=base.RULE_ADMIN_API,
        description='Delete a cluster template from any project.',
        operations=[
            {
                'path': '/v1/clustertemplate/{clustertemplate_ident}',
                'method': 'DELETE'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=CLUSTER_TEMPLATE % 'detail_all_projects',
        check_str=base.RULE_ADMIN_API,
        description=('Retrieve a list of cluster templates with detail across '
                     'projects.'),
        operations=[
            {
                'path': '/v1/clustertemplates',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=CLUSTER_TEMPLATE % 'detail',
        check_str=base.RULE_DENY_CLUSTER_USER,
        description='Retrieve a list of cluster templates with detail.',
        operations=[
            {
                'path': '/v1/clustertemplates',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=CLUSTER_TEMPLATE % 'get',
        check_str=base.RULE_DENY_CLUSTER_USER,
        description='Retrieve information about the given cluster template.',
        operations=[
            {
                'path': '/v1/clustertemplate/{clustertemplate_ident}',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=CLUSTER_TEMPLATE % 'get_one_all_projects',
        check_str=base.RULE_ADMIN_API,
        description=('Retrieve information about the given cluster template '
                     'across project.'),
        operations=[
            {
                'path': '/v1/clustertemplate/{clustertemplate_ident}',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=CLUSTER_TEMPLATE % 'get_all',
        check_str=base.RULE_DENY_CLUSTER_USER,
        description='Retrieve a list of cluster templates.',
        operations=[
            {
                'path': '/v1/clustertemplates',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=CLUSTER_TEMPLATE % 'get_all_all_projects',
        check_str=base.RULE_ADMIN_API,
        description='Retrieve a list of cluster templates across projects.',
        operations=[
            {
                'path': '/v1/clustertemplates',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=CLUSTER_TEMPLATE % 'update',
        check_str=base.RULE_DENY_CLUSTER_USER,
        description='Update an existing cluster template.',
        operations=[
            {
                'path': '/v1/clustertemplate/{clustertemplate_ident}',
                'method': 'PATCH'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=CLUSTER_TEMPLATE % 'update_all_projects',
        check_str=base.RULE_ADMIN_API,
        description='Update an existing cluster template.',
        operations=[
            {
                'path': '/v1/clustertemplate/{clustertemplate_ident}',
                'method': 'PATCH'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=CLUSTER_TEMPLATE % 'publish',
        check_str=base.RULE_ADMIN_API,
        description='Publish an existing cluster template.',
        operations=[
            {
                'path': '/v1/clustertemplates',
                'method': 'POST'
            },
            {
                'path': '/v1/clustertemplates',
                'method': 'PATCH'
            }
        ]
    )
]


def list_rules():
    return rules



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""adding magnum_service functionality

Revision ID: 27ad304554e2
Revises: 1d045384b966
Create Date: 2015-09-01 18:27:14.371860

"""

# revision identifiers, used by Alembic.
revision = '27ad304554e2'
down_revision = '1d045384b966'

from alembic import op  # noqa: E402
import sqlalchemy as sa  # noqa: E402


def upgrade():
    op.create_table(
        'magnum_service',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('report_count', sa.Integer(), nullable=False),
        sa.Column('host', sa.String(length=255), nullable=True),
        sa.Column('binary', sa.String(length=255), nullable=True),
        sa.Column('disabled',
                  sa.Boolean(create_constraint=False),
                  nullable=True),
        sa.Column('disabled_reason', sa.String(length=255), nullable=True),
        # 'last_seen_up' has different purpose than 'updated_at'.
        # 'updated_at' refers to any modification of the entry, which can
        # be administrative too, whereas 'last_seen_up' is more related to
        # magnum_service. Modeled after nova/servicegroup
        sa.Column('last_seen_up', sa.DateTime(), nullable=True),
        sa.Column('forced_down',
                  sa.Boolean(create_constraint=False),
                  nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('host', 'binary',
                            name='uniq_magnum_service0host0binary')
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from magnum.tests.functional import python_client_base as base


class TestFedoraKubernetesIronicAPIs(base.BaseK8sTest):
    cluster_complete_timeout = 3200
    cluster_template_kwargs = {
        "tls_disabled": True,
        "network_driver": 'flannel',
        "volume_driver": None,
        "fixed_subnet": 'private-subnet',
        "server_type": 'bm',
        "docker_storage_driver": 'overlay',
        "labels": {
            "system_pods_initial_delay": 3600,
            "system_pods_timeout": 600,
            "kube_dashboard_enabled": False
        }
    }



# Copyright 2013 - Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Magnum base exception handling.

Includes decorator for re-raising Magnum-type exceptions.

"""

import functools
import sys

from keystoneclient import exceptions as keystone_exceptions
from oslo_config import cfg
from oslo_log import log as logging
import six

import magnum.conf
from magnum.i18n import _


LOG = logging.getLogger(__name__)

CONF = magnum.conf.CONF

try:
    CONF.import_opt('fatal_exception_format_errors',
                    'oslo_versionedobjects.exception')
except cfg.NoSuchOptError:
    # Note:work around for magnum run against master branch
    # in devstack gate job, as magnum not branched yet
    # verisonobjects kilo/master different version can
    # cause issue here. As it changed import group. So
    # add here before branch to prevent gate failure.
    # Bug: #1447873
    CONF.import_opt('fatal_exception_format_errors',
                    'oslo_versionedobjects.exception',
                    group='oslo_versionedobjects')


def wrap_keystone_exception(func):
    """Wrap keystone exceptions and throw Magnum specific exceptions."""
    @functools.wraps(func)
    def wrapped(*args, **kw):
        try:
            return func(*args, **kw)
        except keystone_exceptions.AuthorizationFailure:
            raise AuthorizationFailure(
                client=func.__name__, message="reason: %s" % sys.exc_info()[1])
        except keystone_exceptions.ClientException:
            raise AuthorizationFailure(
                client=func.__name__,
                message="unexpected keystone client error occurred: %s"
                        % sys.exc_info()[1])
    return wrapped


class MagnumException(Exception):
    """Base Magnum Exception

    To correctly use this class, inherit from it and define
    a 'message' property. That message will get printf'd
    with the keyword arguments provided to the constructor.

    """
    message = _("An unknown exception occurred.")
    code = 500

    def __init__(self, message=None, **kwargs):
        self.kwargs = kwargs

        if 'code' not in self.kwargs and hasattr(self, 'code'):
            self.kwargs['code'] = self.code

        if message:
            self.message = message

        try:
            self.message = self.message % kwargs
        except Exception:
            # kwargs doesn't match a variable in the message
            # log the issue and the kwargs
            LOG.exception('Exception in string format operation, '
                          'kwargs: %s', kwargs)
            try:
                if CONF.fatal_exception_format_errors:
                    raise
            except cfg.NoSuchOptError:
                # Note: work around for Bug: #1447873
                if CONF.oslo_versionedobjects.fatal_exception_format_errors:
                    raise

        super(MagnumException, self).__init__(self.message)

    def __str__(self):
        if six.PY3:
            return self.message
        return self.message.encode('utf-8')

    def __unicode__(self):
        return self.message

    def format_message(self):
        if self.__class__.__name__.endswith('_Remote'):
            return self.args[0]
        else:
            return six.text_type(self)


class ObjectNotFound(MagnumException):
    message = _("The %(name)s %(id)s could not be found.")
    code = 404


class ProjectNotFound(ObjectNotFound):
    message = _("The %(name)s %(id)s could not be found.")


class ResourceNotFound(ObjectNotFound):
    message = _("The %(name)s resource %(id)s could not be found.")


class AuthorizationFailure(MagnumException):
    message = _("%(client)s connection failed. %(message)s")
    code = 403


class Invalid(MagnumException):
    message = _("Unacceptable parameters.")
    code = 400


class InvalidUUID(Invalid):
    message = _("Expected a uuid but received %(uuid)s.")


class InvalidName(Invalid):
    message = _("Expected a name but received %(name)s.")


class InvalidDiscoveryURL(Invalid):
    message = _("Received invalid discovery URL '%(discovery_url)s' for "
                "discovery endpoint '%(discovery_endpoint)s'.")


class GetDiscoveryUrlFailed(MagnumException):
    message = _("Failed to get discovery url from '%(discovery_endpoint)s'.")


class InvalidClusterDiscoveryURL(Invalid):
    message = _("Invalid discovery URL '%(discovery_url)s'.")


class InvalidClusterSize(Invalid):
    message = _("Expected cluster size %(expect_size)d but get cluster "
                "size %(size)d from '%(discovery_url)s'.")


class GetClusterSizeFailed(MagnumException):
    message = _("Failed to get the size of cluster from '%(discovery_url)s'.")


class InvalidIdentity(Invalid):
    message = _("Expected an uuid or int but received %(identity)s.")


class InvalidCsr(Invalid):
    message = _("Received invalid csr %(csr)s.")


class InvalidSubnet(Invalid):
    message = _("Received invalid subnet %(subnet)s.")


class InvalidVersion(Invalid):
    message = _("Received invalid tag for %(tag)s.")


class HTTPNotFound(ResourceNotFound):
    pass


class Conflict(MagnumException):
    message = _('Conflict.')
    code = 409


class ApiVersionsIntersect(Invalid):
    message = _("Version of %(name)s %(min_ver)s %(max_ver)s intersects "
                "with another versions.")


# Cannot be templated as the error syntax varies.
# msg needs to be constructed when raised.
class InvalidParameterValue(Invalid):
    message = _("%(err)s")


class PatchError(Invalid):
    message = _("Couldn't apply patch '%(patch)s'. Reason: %(reason)s")


class NotAuthorized(MagnumException):
    message = _("Not authorized.")
    code = 403


class PolicyNotAuthorized(NotAuthorized):
    message = _("Policy doesn't allow %(action)s to be performed.")


class InvalidMAC(Invalid):
    message = _("Expected a MAC address but received %(mac)s.")


class InvalidDNS(Invalid):
    message = _(
        "Expected a single dns address or comma separated dns list, "
        "but received %(dns)s.")


class ConfigInvalid(Invalid):
    message = _("Invalid configuration file. %(error_msg)s")


class ClusterTemplateNotFound(ResourceNotFound):
    message = _("ClusterTemplate %(clustertemplate)s could not be found.")


class ClusterTemplateAlreadyExists(Conflict):
    message = _("A ClusterTemplate with UUID %(uuid)s already exists.")


class ClusterTemplateReferenced(Invalid):
    message = _("ClusterTemplate %(clustertemplate)s is referenced by one or"
                " multiple clusters.")


class ClusterTemplatePublishDenied(NotAuthorized):
    message = _("Not authorized to set public or hidden flag for cluster"
                " template.")


class ClusterNotFound(ResourceNotFound):
    message = _("Cluster %(cluster)s could not be found.")


class ClusterAlreadyExists(Conflict):
    message = _("A cluster with UUID %(uuid)s already exists.")


class NotSupported(MagnumException):
    message = _("%(operation)s is not supported.")
    code = 400


class ClusterTypeNotSupported(NotSupported):
    message = _("Cluster type (%(server_type)s, %(os)s, %(coe)s)"
                " not supported.")


class RequiredParameterNotProvided(Invalid):
    message = _("Required parameter %(heat_param)s not provided.")


class OperationInProgress(Invalid):
    message = _("Cluster %(cluster_name)s already has an operation in "
                "progress.")


class VolumeTypeNotFound(ResourceNotFound):
    """The code here changed to 400 according to the latest document."""
    message = _("Valid volume type could not be found.")
    code = 400


class ImageNotFound(ResourceNotFound):
    """The code here changed to 400 according to the latest document."""
    message = _("Image %(image_id)s could not be found.")
    code = 400


class ImageNotAuthorized(NotAuthorized):
    message = _("Not authorized for image %(image_id)s.")


class OSDistroFieldNotFound(ResourceNotFound):
    """The code here changed to 400 according to the latest document."""
    message = _("Image %(image_id)s doesn't contain os_distro field.")
    code = 400


class X509KeyPairNotFound(ResourceNotFound):
    message = _("A key pair %(x509keypair)s could not be found.")


class X509KeyPairAlreadyExists(Conflict):
    message = _("A key pair with UUID %(uuid)s already exists.")


class CertificateStorageException(MagnumException):
    message = _("Could not store certificate: %(msg)s")


class CertificateValidationError(Invalid):
    message = _("Extension '%(extension)s' not allowed")


class KeyPairNotFound(ResourceNotFound):
    message = _("Unable to find keypair %(keypair)s.")


class MagnumServiceNotFound(ResourceNotFound):
    message = _("A magnum service %(magnum_service_id)s could not be found.")


class MagnumServiceAlreadyExists(Conflict):
    message = _("A magnum service with ID %(id)s already exists.")


class UnsupportedK8sQuantityFormat(Invalid):
    message = _("Unsupported quantity format for k8s cluster.")


class UnsupportedDockerQuantityFormat(Invalid):
    message = _("Unsupported quantity format for Swarm cluster.")


class FlavorNotFound(ResourceNotFound):
    """The code here changed to 400 according to the latest document."""
    message = _("Unable to find flavor %(flavor)s.")
    code = 400


class FixedNetworkNotFound(ResourceNotFound):
    """The code here changed to 400 according to the latest document."""
    """"Ensure the network is private."""
    message = _("Unable to find fixed network %(network)s.")
    code = 400


class FixedSubnetNotFound(ResourceNotFound):
    """The code here changed to 400 according to the latest document."""
    message = _("Unable to find fixed subnet %(subnet)s.")
    code = 400


class ExternalNetworkNotFound(ResourceNotFound):
    """The code here changed to 400 according to the latest document."""
    """"Ensure the network is not private."""
    message = _("Unable to find external network %(network)s.")
    code = 400


class TrustCreateFailed(MagnumException):
    message = _("Failed to create trust for trustee %(trustee_user_id)s.")


class TrustDeleteFailed(MagnumException):
    message = _("Failed to delete trust %(trust_id)s.")


class TrusteeCreateFailed(MagnumException):
    message = _("Failed to create trustee %(username)s "
                "in domain %(domain_id)s")


class TrusteeDeleteFailed(MagnumException):
    message = _("Failed to delete trustee %(trustee_id)s")


class QuotaAlreadyExists(Conflict):
    message = _("Quota for project %(project_id)s already exists "
                "for resource %(resource)s.")


class QuotaNotFound(ResourceNotFound):
    message = _("Quota could not be found: %(msg)s")


class ResourceLimitExceeded(NotAuthorized):
    message = _('Resource limit exceeded: %(msg)s')


class RegionsListFailed(MagnumException):
    message = _("Failed to list regions.")


class ServicesListFailed(MagnumException):
    message = _("Failed to list services.")


class TrusteeOrTrustToClusterFailed(MagnumException):
    message = _("Failed to create trustee or trust for Cluster: "
                "%(cluster_uuid)s")


class CertificatesToClusterFailed(MagnumException):
    message = _("Failed to create certificates for Cluster: %(cluster_uuid)s")


class FederationNotFound(ResourceNotFound):
    message = _("Federation %(federation)s could not be found.")


class FederationAlreadyExists(Conflict):
    message = _("A federation with UUID %(uuid)s already exists.")


class MemberAlreadyExists(Conflict):
    message = _("A cluster with UUID %(uuid)s is already a member of the "
                "federation %(federation_name)s.")


class PreDeletionFailed(Conflict):
    message = _("Failed to pre-delete resources for cluster %(cluster_uuid)s, "
                "error: %(msg)s.")


class NodeGroupAlreadyExists(Conflict):
    message = _("A node group with name %(name)s already exists in the "
                "cluster %(cluster_id)s.")


class NodeGroupNotFound(ResourceNotFound):
    message = _("Nodegroup %(nodegroup)s could not be found.")


class MasterNGResizeNotSupported(NotSupported):
    message = _("Resizing a master nodegroup is not supported.")


class NGResizeOutBounds(Invalid):
    message = _("Resizing %(nodegroup)s outside the allowed range: "
                "min_node_count = %(min_nc)s, "
                "max_node_count = %(max_nc)s")


class DeletingDefaultNGNotSupported(NotSupported):
    message = _("Deleting a default nodegroup is not supported.")


class NodeGroupInvalidInput(Conflict):
    message = _("%(attr)s for %(nodegroup)s is invalid (%(expl)s).")


class CreateMasterNodeGroup(NotSupported):
    message = _("Creating master nodegroups is currently not supported.")


class NgOperationInProgress(Invalid):
    message = _("Nodegroup %(nodegroup)s already has an operation in "
                "progress.")


class InvalidClusterTemplateForUpgrade(Conflict):
    message = _("Cluster Template is not valid for upgrade: %(reason)s")


class ClusterAPIAddressUnavailable(Conflict):
    message = _("Cluster API address is not available yet")



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from magnum.tests.functional.api.v1.models import baymodel_model
from magnum.tests.functional.common import client


class BayModelClient(client.MagnumClient):
    """Encapsulates REST calls and maps JSON to/from models"""

    @classmethod
    def baymodels_uri(cls, filters=None):
        """Construct baymodels uri with optional filters

        :param filters: Optional k:v dict that's converted to url query
        :returns: url string
        """

        url = "/baymodels"
        if filters:
            url = cls.add_filters(url, filters)
        return url

    @classmethod
    def baymodel_uri(cls, baymodel_id):
        """Construct baymodel uri

        :param baymodel_id: baymodel uuid or name
        :returns: url string
        """

        return "{0}/{1}".format(cls.baymodels_uri(), baymodel_id)

    def list_baymodels(self, filters=None, **kwargs):
        """Makes GET /baymodels request and returns BayModelCollection

        Abstracts REST call to return all baymodels

        :param filters: Optional k:v dict that's converted to url query
        :returns: response object and BayModelCollection object
        """

        resp, body = self.get(self.baymodels_uri(filters), **kwargs)
        return self.deserialize(resp, body, baymodel_model.BayModelCollection)

    def get_baymodel(self, baymodel_id, **kwargs):
        """Makes GET /baymodel request and returns BayModelEntity

        Abstracts REST call to return a single baymodel based on uuid or name

        :param baymodel_id: baymodel uuid or name
        :returns: response object and BayModelCollection object
        """

        resp, body = self.get(self.baymodel_uri(baymodel_id))
        return self.deserialize(resp, body, baymodel_model.BayModelEntity)

    def post_baymodel(self, model, **kwargs):
        """Makes POST /baymodel request and returns BayModelEntity

        Abstracts REST call to create new baymodel

        :param model: BayModelEntity
        :returns: response object and BayModelEntity object
        """

        resp, body = self.post(
            self.baymodels_uri(),
            body=model.to_json(), **kwargs)
        return self.deserialize(resp, body, baymodel_model.BayModelEntity)

    def patch_baymodel(self, baymodel_id, baymodelpatch_listmodel, **kwargs):
        """Makes PATCH /baymodel request and returns BayModelEntity

        Abstracts REST call to update baymodel attributes

        :param baymodel_id: UUID of baymodel
        :param baymodelpatch_listmodel: BayModelPatchCollection
        :returns: response object and BayModelEntity object
        """

        resp, body = self.patch(
            self.baymodel_uri(baymodel_id),
            body=baymodelpatch_listmodel.to_json(), **kwargs)
        return self.deserialize(resp, body, baymodel_model.BayModelEntity)

    def delete_baymodel(self, baymodel_id, **kwargs):
        """Makes DELETE /baymodel request and returns response object

        Abstracts REST call to delete baymodel based on uuid or name

        :param baymodel_id: UUID or name of baymodel
        :returns: response object
        """

        return self.delete(self.baymodel_uri(baymodel_id), **kwargs)



# Copyright 2015 Huawei Technologies Co.,LTD.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock
from unittest.mock import patch

from magnum.conductor import utils
from magnum import objects
from magnum.tests import base


class TestConductorUtils(base.TestCase):

    def _test_retrieve_cluster(self, expected_cluster_uuid,
                               mock_cluster_get_by_uuid):
        expected_context = 'context'
        utils.retrieve_cluster(expected_context, expected_cluster_uuid)
        mock_cluster_get_by_uuid.assert_called_once_with(
            expected_context, expected_cluster_uuid)

    def get_fake_id(self):
        return '5d12f6fd-a196-4bf0-ae4c-1f639a523a52'

    def _get_type_uri(self):
        return 'service/security/account/user'

    @patch('magnum.objects.ClusterTemplate.get_by_uuid')
    def test_retrieve_cluster_template(self,
                                       mock_cluster_template_get_by_uuid):
        expected_context = 'context'
        expected_cluster_template_uuid = 'ClusterTemplate_uuid'

        cluster = objects.Cluster({})
        cluster.cluster_template_id = expected_cluster_template_uuid

        utils.retrieve_cluster_template(expected_context, cluster)

        mock_cluster_template_get_by_uuid.assert_called_once_with(
            expected_context,
            expected_cluster_template_uuid)

    @patch('oslo_utils.uuidutils.is_uuid_like')
    @patch('magnum.objects.Cluster.get_by_name')
    def test_retrieve_cluster_uuid_from_name(self, mock_cluster_get_by_name,
                                             mock_uuid_like):
        cluster = objects.Cluster(uuid='5d12f6fd-a196-4bf0-ae4c-1f639a523a52')
        mock_uuid_like.return_value = False
        mock_cluster_get_by_name.return_value = cluster
        cluster_uuid = utils.retrieve_cluster_uuid('context', 'fake_name')
        self.assertEqual('5d12f6fd-a196-4bf0-ae4c-1f639a523a52', cluster_uuid)

        mock_uuid_like.assert_called_once_with('fake_name')
        mock_cluster_get_by_name.assert_called_once_with('context',
                                                         'fake_name')

    @patch('oslo_utils.uuidutils.is_uuid_like')
    @patch('magnum.objects.Cluster.get_by_name')
    def test_retrieve_cluster_uuid_from_uuid(self, mock_cluster_get_by_name,
                                             mock_uuid_like):
        cluster_uuid = utils.retrieve_cluster_uuid(
            'context',
            '5d12f6fd-a196-4bf0-ae4c-1f639a523a52')
        self.assertEqual('5d12f6fd-a196-4bf0-ae4c-1f639a523a52', cluster_uuid)
        mock_uuid_like.return_value = True
        mock_cluster_get_by_name.assert_not_called()

    def _get_heat_stacks_get_mock_obj(self, status):
        mock_stack = mock.MagicMock()
        mock_osc = mock.MagicMock()
        mock_stack_obj = mock.MagicMock()
        mock_stack_obj.stack_status = status
        stack_get = mock.MagicMock()
        stack_get.get.return_value = mock_stack_obj
        mock_stack.stacks = stack_get
        mock_osc.heat.return_value = mock_stack
        return mock_osc

    @patch('magnum.conductor.utils.retrieve_cluster')
    @patch('magnum.conductor.utils.clients.OpenStackClients')
    def test_object_has_stack_invalid_status(self, mock_oscs,
                                             mock_retrieve_cluster):

        mock_osc = self._get_heat_stacks_get_mock_obj("INVALID_STATUS")
        mock_oscs.return_value = mock_osc
        self.assertTrue(utils.object_has_stack('context', self.get_fake_id()))
        mock_retrieve_cluster.assert_called_with('context', self.get_fake_id())

    @patch('magnum.conductor.utils.retrieve_cluster')
    @patch('magnum.conductor.utils.clients.OpenStackClients')
    def test_object_has_stack_delete_in_progress(self, mock_oscs,
                                                 mock_retrieve_cluster):

        mock_osc = self._get_heat_stacks_get_mock_obj("DELETE_IN_PROGRESS")
        mock_oscs.return_value = mock_osc
        self.assertFalse(utils.object_has_stack('context', self.get_fake_id()))
        mock_retrieve_cluster.assert_called_with('context', self.get_fake_id())

    @patch('magnum.conductor.utils.retrieve_cluster')
    @patch('magnum.conductor.utils.clients.OpenStackClients')
    def test_object_has_stack_delete_complete_status(self, mock_oscs,
                                                     mock_retrieve_cluster):
        mock_osc = self._get_heat_stacks_get_mock_obj("DELETE_COMPLETE")
        mock_oscs.return_value = mock_osc
        self.assertFalse(utils.object_has_stack('context', self.get_fake_id()))
        mock_retrieve_cluster.assert_called_with('context', self.get_fake_id())

    @patch('magnum.objects.Cluster.get_by_uuid')
    def test_retrieve_cluster_uuid(self, mock_get_by_uuid):
        mock_get_by_uuid.return_value = True
        utils.retrieve_cluster('context',
                               '5d12f6fd-a196-4bf0-ae4c-1f639a523a52')
        self.assertTrue(mock_get_by_uuid.called)

    @patch('magnum.objects.Cluster.get_by_name')
    def test_retrieve_cluster_name(self, mock_get_by_name):
        mock_get_by_name.return_value = mock.MagicMock()
        utils.retrieve_cluster('context', '1')
        self.assertTrue(mock_get_by_name.called)

    @patch('magnum.conductor.utils.resource.Resource')
    def test_get_request_audit_info_with_none_context(self, mock_resource):
        mock_resource.return_value = 'resource'
        result = utils._get_request_audit_info(context=None)
        self.assertTrue(mock_resource.called)
        self.assertEqual(result, 'resource')

    def _assert_for_user_project_domain_resource(self, result, ctxt, mock_res):
        mock_res.assert_called_once_with(typeURI=self._get_type_uri())
        self.assertEqual(result.user_id, ctxt.user_id)
        self.assertEqual(result.project_id, ctxt.project_id)
        self.assertEqual(result.domain_id, ctxt.domain_id)

    def _get_context(self, user_id=None, project_id=None, domain_id=None):
        context = self.mock_make_context()
        context.user_id = user_id
        context.project_id = project_id
        context.domain_id = domain_id
        return context

    @patch('magnum.conductor.utils.resource.Resource')
    def test_get_request_audit_info_with_none_userid(self, mock_resource):
        context = self._get_context(project_id='test_project_id',
                                    domain_id='test_domain_id')

        mock_resource.return_value = context
        result = utils._get_request_audit_info(context)
        self._assert_for_user_project_domain_resource(result, context,
                                                      mock_resource)

    @patch('magnum.conductor.utils.resource.Resource')
    def test_get_request_audit_info_with_none_projectid(self, mock_resource):
        context = self._get_context(user_id='test_user_id',
                                    domain_id='test_domain_id')

        mock_resource.return_value = context
        result = utils._get_request_audit_info(context)
        self._assert_for_user_project_domain_resource(result, context,
                                                      mock_resource)

    @patch('magnum.conductor.utils.resource.Resource')
    def test_get_request_audit_info_with_none_domainid(self, mock_resource):
        context = self._get_context(user_id='test_user_id',
                                    project_id='test_project_id')

        mock_resource.return_value = context
        result = utils._get_request_audit_info(context)
        self._assert_for_user_project_domain_resource(result, context,
                                                      mock_resource)

    @patch('magnum.conductor.utils.resource.Resource')
    def test_get_request_audit_info_with_none_domainid_userid(self,
                                                              mock_resource):

        context = self._get_context(project_id='test_project_id')
        mock_resource.return_value = context
        result = utils._get_request_audit_info(context)
        self._assert_for_user_project_domain_resource(result, context,
                                                      mock_resource)

    @patch('magnum.conductor.utils.resource.Resource')
    def test_get_request_audit_info_with_none_userid_projectid(self,
                                                               mock_resource):

        context = self._get_context(domain_id='test_domain_id')
        mock_resource.return_value = context
        result = utils._get_request_audit_info(context)
        self._assert_for_user_project_domain_resource(result, context,
                                                      mock_resource)

    @patch('magnum.conductor.utils.resource.Resource')
    def test_get_request_audit_info_with_none_domain_project_id(self,
                                                                mock_resource):

        context = self._get_context(user_id='test_user_id')
        mock_resource.return_value = context
        result = utils._get_request_audit_info(context)
        self._assert_for_user_project_domain_resource(result, context,
                                                      mock_resource)



# Copyright 2016 - Fujitsu, Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

import six

from magnum.cmd import db_manage
from magnum.tests import base


class TestMagnumDbManage(base.TestCase):

    def setUp(self):
        super(TestMagnumDbManage, self).setUp()

        def clear_conf():
            db_manage.CONF.reset()
            db_manage.CONF.unregister_opt(db_manage.command_opt)

        clear_conf()
        self.addCleanup(clear_conf)

    @mock.patch('magnum.db.migration.version')
    @mock.patch('sys.argv', ['magnum-db-manage', 'version'])
    def test_db_manage_version(self, mock_version):
        with mock.patch('sys.stdout', new=six.StringIO()) as fakeOutput:
            mock_version.return_value = '123456'
            db_manage.main()
            self.assertEqual('Current DB revision is 123456\n',
                             fakeOutput.getvalue())
            mock_version.assert_called_once_with()

    @mock.patch('magnum.db.migration.upgrade')
    @mock.patch('sys.argv', ['magnum-db-manage', 'upgrade'])
    def test_db_manage_upgrade(self, mock_upgrade):
        db_manage.main()
        mock_upgrade.assert_called_once_with(base.CONF.command.revision)

    @mock.patch('magnum.db.migration.stamp')
    @mock.patch('sys.argv', ['magnum-db-manage', 'stamp', 'foo bar'])
    def test_db_manage_stamp(self, mock_stamp):
        db_manage.main()
        mock_stamp.assert_called_once_with('foo bar')

    @mock.patch('magnum.db.migration.revision')
    @mock.patch('sys.argv', ['magnum-db-manage', 'revision', '-m', 'foo bar'])
    def test_db_manage_revision(self, mock_revision):
        db_manage.main()
        mock_revision.assert_called_once_with(
            message='foo bar',
            autogenerate=base.CONF.command.autogenerate)



# Copyright 2015 NEC Corporation.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import enum


class Extensions(enum.Enum):
    __order__ = ('AUTHORITY_KEY_IDENTIFIER SUBJECT_KEY_IDENTIFIER '
                 'AUTHORITY_INFORMATION_ACCESS BASIC_CONSTRAINTS '
                 'CRL_DISTRIBUTION_POINTS CERTIFICATE_POLICIES '
                 'EXTENDED_KEY_USAGE OCSP_NO_CHECK INHIBIT_ANY_POLICY '
                 'KEY_USAGE NAME_CONSTRAINTS SUBJECT_ALTERNATIVE_NAME '
                 'ISSUER_ALTERNATIVE_NAME')

    AUTHORITY_KEY_IDENTIFIER = "authorityKeyIdentifier"
    SUBJECT_KEY_IDENTIFIER = "subjectKeyIdentifier"
    AUTHORITY_INFORMATION_ACCESS = "authorityInfoAccess"
    BASIC_CONSTRAINTS = "basicConstraints"
    CRL_DISTRIBUTION_POINTS = "cRLDistributionPoints"
    CERTIFICATE_POLICIES = "certificatePolicies"
    EXTENDED_KEY_USAGE = "extendedKeyUsage"
    OCSP_NO_CHECK = "OCSPNoCheck"
    INHIBIT_ANY_POLICY = "inhibitAnyPolicy"
    KEY_USAGE = "keyUsage"
    NAME_CONSTRAINTS = "nameConstraints"
    SUBJECT_ALTERNATIVE_NAME = "subjectAltName"
    ISSUER_ALTERNATIVE_NAME = "issuerAltName"


class KeyUsages(enum.Enum):
    __order__ = ('DIGITAL_SIGNATURE CONTENT_COMMITMENT KEY_ENCIPHERMENT '
                 'DATA_ENCIPHERMENT KEY_AGREEMENT KEY_CERT_SIGN '
                 'CRL_SIGN ENCIPHER_ONLY DECIPHER_ONLY')

    DIGITAL_SIGNATURE = ("Digital Signature", "digital_signature")
    CONTENT_COMMITMENT = ("Non Repudiation", "content_commitment")
    KEY_ENCIPHERMENT = ("Key Encipherment", "key_encipherment")
    DATA_ENCIPHERMENT = ("Data Encipherment", "data_encipherment")
    KEY_AGREEMENT = ("Key Agreement", "key_agreement")
    KEY_CERT_SIGN = ("Certificate Sign", "key_cert_sign")
    CRL_SIGN = ("CRL Sign", "crl_sign")
    ENCIPHER_ONLY = ("Encipher Only", "encipher_only")
    DECIPHER_ONLY = ("Decipher Only", "decipher_only")



# Copyright 2014 - Rackspace Hosting.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from barbicanclient.v1 import client as barbicanclient
from cinderclient.v2 import client as cinder_client
from glanceclient import client as glanceclient
from heatclient import client as heatclient
from keystoneauth1.exceptions import catalog
from neutronclient.v2_0 import client as neutronclient
from novaclient import client as novaclient
from octaviaclient.api.v2 import octavia
from oslo_log import log as logging

from magnum.common import exception
from magnum.common import keystone
import magnum.conf

CONF = magnum.conf.CONF
LOG = logging.getLogger(__name__)


class OpenStackClients(object):
    """Convenience class to create and cache client instances."""

    def __init__(self, context):
        self.context = context
        self._keystone = None
        self._heat = None
        self._glance = None
        self._barbican = None
        self._nova = None
        self._neutron = None
        self._octavia = None
        self._cinder = None

    def url_for(self, **kwargs):
        return self.keystone().session.get_endpoint(**kwargs)

    def magnum_url(self):
        endpoint_type = self._get_client_option('magnum', 'endpoint_type')
        region_name = self._get_client_option('magnum', 'region_name')
        try:
            return self.url_for(service_type='container-infra',
                                interface=endpoint_type,
                                region_name=region_name)
        except catalog.EndpointNotFound:
            url = self.url_for(service_type='container',
                               interface=endpoint_type,
                               region_name=region_name)
            LOG.warning('Service type "container" is deprecated and will '
                        'be removed in a subsequent release')
            return url

    def cinder_region_name(self):
        cinder_region_name = self._get_client_option('cinder', 'region_name')
        return self.keystone().get_validate_region_name(cinder_region_name)

    @property
    def auth_url(self):
        return self.keystone().auth_url

    @property
    def auth_token(self):
        return self.context.auth_token or self.keystone().auth_token

    def keystone(self):
        if self._keystone:
            return self._keystone

        self._keystone = keystone.KeystoneClientV3(self.context)
        return self._keystone

    def _get_client_option(self, client, option):
        return getattr(getattr(CONF, '%s_client' % client), option)

    @exception.wrap_keystone_exception
    def octavia(self):
        if self._octavia:
            return self._octavia

        region_name = self._get_client_option('octavia', 'region_name')
        endpoint_type = self._get_client_option('octavia', 'endpoint_type')
        endpoint = self.url_for(service_type='load-balancer',
                                interface=endpoint_type,
                                region_name=region_name)
        session = self.keystone().session
        return octavia.OctaviaAPI(session=session,
                                  service_type='load-balancer',
                                  endpoint=endpoint)

    @exception.wrap_keystone_exception
    def heat(self):
        if self._heat:
            return self._heat

        endpoint_type = self._get_client_option('heat', 'endpoint_type')
        region_name = self._get_client_option('heat', 'region_name')
        heatclient_version = self._get_client_option('heat', 'api_version')
        endpoint = self.url_for(service_type='orchestration',
                                interface=endpoint_type,
                                region_name=region_name)

        args = {
            'endpoint': endpoint,
            'auth_url': self.auth_url,
            'token': self.auth_token,
            'username': None,
            'password': None,
            'ca_file': self._get_client_option('heat', 'ca_file'),
            'cert_file': self._get_client_option('heat', 'cert_file'),
            'key_file': self._get_client_option('heat', 'key_file'),
            'insecure': self._get_client_option('heat', 'insecure')
        }
        self._heat = heatclient.Client(heatclient_version, **args)

        return self._heat

    @exception.wrap_keystone_exception
    def glance(self):
        if self._glance:
            return self._glance

        endpoint_type = self._get_client_option('glance', 'endpoint_type')
        region_name = self._get_client_option('glance', 'region_name')
        glanceclient_version = self._get_client_option('glance', 'api_version')
        endpoint = self.url_for(service_type='image',
                                interface=endpoint_type,
                                region_name=region_name)
        args = {
            'endpoint': endpoint,
            'auth_url': self.auth_url,
            'token': self.auth_token,
            'username': None,
            'password': None,
            'cacert': self._get_client_option('glance', 'ca_file'),
            'cert': self._get_client_option('glance', 'cert_file'),
            'key': self._get_client_option('glance', 'key_file'),
            'insecure': self._get_client_option('glance', 'insecure')
        }
        self._glance = glanceclient.Client(glanceclient_version, **args)

        return self._glance

    @exception.wrap_keystone_exception
    def barbican(self):
        if self._barbican:
            return self._barbican

        endpoint_type = self._get_client_option('barbican', 'endpoint_type')
        region_name = self._get_client_option('barbican', 'region_name')
        endpoint = self.url_for(service_type='key-manager',
                                interface=endpoint_type,
                                region_name=region_name)
        session = self.keystone().session
        self._barbican = barbicanclient.Client(session=session,
                                               endpoint=endpoint)

        return self._barbican

    @exception.wrap_keystone_exception
    def nova(self):
        if self._nova:
            return self._nova
        endpoint_type = self._get_client_option('nova', 'endpoint_type')
        region_name = self._get_client_option('nova', 'region_name')
        novaclient_version = self._get_client_option('nova', 'api_version')
        endpoint = self.url_for(service_type='compute',
                                interface=endpoint_type,
                                region_name=region_name)
        args = {
            'cacert': self._get_client_option('nova', 'ca_file'),
            'insecure': self._get_client_option('nova', 'insecure')
        }

        session = self.keystone().session
        self._nova = novaclient.Client(novaclient_version,
                                       session=session,
                                       endpoint_override=endpoint, **args)
        return self._nova

    @exception.wrap_keystone_exception
    def neutron(self):
        if self._neutron:
            return self._neutron
        endpoint_type = self._get_client_option('neutron', 'endpoint_type')
        region_name = self._get_client_option('neutron', 'region_name')
        endpoint = self.url_for(service_type='network',
                                interface=endpoint_type,
                                region_name=region_name)

        args = {
            'auth_url': self.auth_url,
            'token': self.auth_token,
            'endpoint_url': endpoint,
            'endpoint_type': endpoint_type,
            'ca_cert': self._get_client_option('neutron', 'ca_file'),
            'insecure': self._get_client_option('neutron', 'insecure')
        }
        self._neutron = neutronclient.Client(**args)
        return self._neutron

    @exception.wrap_keystone_exception
    def cinder(self):
        if self._cinder:
            return self._cinder
        endpoint_type = self._get_client_option('cinder', 'endpoint_type')
        region_name = self._get_client_option('cinder', 'region_name')
        cinderclient_version = self._get_client_option('cinder', 'api_version')
        endpoint = self.url_for(service_type='block-storage',
                                interface=endpoint_type,
                                region_name=region_name)
        args = {
            'cacert': self._get_client_option('cinder', 'ca_file'),
            'insecure': self._get_client_option('cinder', 'insecure')
        }

        session = self.keystone().session
        self._cinder = cinder_client.Client(cinderclient_version,
                                            session=session,
                                            endpoint_override=endpoint, **args)
        return self._cinder



#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_serialization import jsonutils

from magnum.common import urlfetch
from magnum.conductor import monitors


class DcosMonitor(monitors.MonitorBase):

    def __init__(self, context, cluster):
        super(DcosMonitor, self).__init__(context, cluster)
        self.data = {}

    @property
    def metrics_spec(self):
        return {
            'memory_util': {
                'unit': '%',
                'func': 'compute_memory_util',
            },
            'cpu_util': {
                'unit': '%',
                'func': 'compute_cpu_util',
            },
        }

    # See https://github.com/dcos/adminrouter#ports-summary
    # Use http://<mesos-master>/mesos/ instead of http://<mesos-master>:5050
    def _build_url(self, url, protocol='http', server_name='mesos', path='/'):
        return protocol + '://' + url + '/' + server_name + path

    def _is_leader(self, state):
        return state['leader'] == state['pid']

    def pull_data(self):
        self.data['mem_total'] = 0
        self.data['mem_used'] = 0
        self.data['cpu_total'] = 0
        self.data['cpu_used'] = 0
        for master_addr in self.cluster.master_addresses:
            mesos_master_url = self._build_url(master_addr,
                                               server_name='mesos',
                                               path='/state')
            master = jsonutils.loads(urlfetch.get(mesos_master_url))
            if self._is_leader(master):
                for slave in master['slaves']:
                    self.data['mem_total'] += slave['resources']['mem']
                    self.data['mem_used'] += slave['used_resources']['mem']
                    self.data['cpu_total'] += slave['resources']['cpus']
                    self.data['cpu_used'] += slave['used_resources']['cpus']
                break

    def compute_memory_util(self):
        if self.data['mem_total'] == 0 or self.data['mem_used'] == 0:
            return 0
        else:
            return self.data['mem_used'] * 100 / self.data['mem_total']

    def compute_cpu_util(self):
        if self.data['cpu_total'] == 0 or self.data['cpu_used'] == 0:
            return 0
        else:
            return self.data['cpu_used'] * 100 / self.data['cpu_total']



# coding=utf-8
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_versionedobjects import fields

from magnum.db import api as dbapi
from magnum.objects import base


@base.MagnumObjectRegistry.register
class Stats(base.MagnumObject, base.MagnumObjectDictCompat):
    # Version 1.0: Initial version

    VERSION = '1.0'

    dbapi = dbapi.get_instance()

    fields = {
        'clusters': fields.IntegerField(),
        'nodes': fields.IntegerField(nullable=True)
    }

    @base.remotable_classmethod
    def get_cluster_stats(cls, context, project_id=None):
        """Return cluster stats for the given project.

        :param context: The security context
        :param project_id: project id
        """
        clusters, nodes = cls.dbapi.get_cluster_stats(context, project_id)
        return cls(clusters=clusters, nodes=nodes)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from magnum.tests.functional.common import models


class CertData(models.BaseModel):
    """Data that encapsulates  cert attributes"""
    pass


class CertEntity(models.EntityModel):
    """Entity Model that represents a single instance of CertData"""
    ENTITY_NAME = 'certificate'
    MODEL_TYPE = CertData



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import timeutils
import pecan
import wsme
from wsme import types as wtypes

from magnum.api import attr_validator
from magnum.api.controllers import base
from magnum.api.controllers import link
from magnum.api.controllers.v1 import collection
from magnum.api.controllers.v1 import types
from magnum.api import expose
from magnum.api import utils as api_utils
from magnum.api import validation
from magnum.common import clients
from magnum.common import exception
from magnum.common import name_generator
from magnum.common import policy
from magnum import objects
from magnum.objects import fields


class BayModel(base.APIBase):
    """API representation of a Baymodel.

    This class enforces type checking and value constraints, and converts
    between the internal object model and the API representation of a Baymodel.
    """

    uuid = types.uuid
    """Unique UUID for this Baymodel"""

    name = wtypes.StringType(min_length=1, max_length=255)
    """The name of the Baymodel"""

    coe = wtypes.Enum(wtypes.text, *fields.ClusterType.ALL, mandatory=True)
    """The Container Orchestration Engine for this bay model"""

    image_id = wsme.wsattr(wtypes.StringType(min_length=1, max_length=255),
                           mandatory=True)
    """The image name or UUID to use as a base image for this Baymodel"""

    flavor_id = wtypes.StringType(min_length=1, max_length=255)
    """The flavor of this Baymodel"""

    master_flavor_id = wtypes.StringType(min_length=1, max_length=255)
    """The flavor of the master node for this Baymodel"""

    dns_nameserver = wtypes.IPv4AddressType()
    """The DNS nameserver address"""

    keypair_id = wsme.wsattr(wtypes.StringType(min_length=1, max_length=255),
                             mandatory=True)
    """The name of the nova ssh keypair"""

    external_network_id = wtypes.StringType(min_length=1, max_length=255)
    """The external network to attach to the Bay"""

    fixed_network = wtypes.StringType(min_length=1, max_length=255)
    """The fixed network name to attach to the Bay"""

    fixed_subnet = wtypes.StringType(min_length=1, max_length=255)
    """The fixed subnet name to attach to the Bay"""

    network_driver = wtypes.StringType(min_length=1, max_length=255)
    """The name of the driver used for instantiating container networks"""

    apiserver_port = wtypes.IntegerType(minimum=1024, maximum=65535)
    """The API server port for k8s"""

    docker_volume_size = wtypes.IntegerType(minimum=1)
    """The size in GB of the docker volume"""

    cluster_distro = wtypes.StringType(min_length=1, max_length=255)
    """The Cluster distro for the bay, e.g. coreos, fedora-atomic, etc."""

    links = wsme.wsattr([link.Link], readonly=True)
    """A list containing a self link and associated Baymodel links"""

    http_proxy = wtypes.StringType(min_length=1, max_length=255)
    """Address of a proxy that will receive all HTTP requests and relay them.
       The format is a URL including a port number.
       """

    https_proxy = wtypes.StringType(min_length=1, max_length=255)
    """Address of a proxy that will receive all HTTPS requests and relay them.
       The format is a URL including a port number.
       """

    no_proxy = wtypes.StringType(min_length=1, max_length=255)
    """A comma separated list of IPs for which proxies should not be
       used in the bay
       """

    volume_driver = wtypes.StringType(min_length=1, max_length=255)
    """The name of the driver used for instantiating container volumes"""

    registry_enabled = wsme.wsattr(types.boolean, default=False)
    """Indicates whether the docker registry is enabled"""

    labels = wtypes.DictType(wtypes.text, wtypes.text)
    """One or more key/value pairs"""

    tls_disabled = wsme.wsattr(types.boolean, default=False)
    """Indicates whether TLS should be disabled"""

    public = wsme.wsattr(types.boolean, default=False)
    """Indicates whether the Baymodel is public or not."""

    server_type = wsme.wsattr(wtypes.Enum(wtypes.text, *fields.ServerType.ALL),
                              default='vm')
    """Server type for this bay model"""

    insecure_registry = wtypes.StringType(min_length=1, max_length=255)
    """Insecure registry URL when creating a Baymodel"""

    docker_storage_driver = wtypes.StringType(min_length=1, max_length=255)
    """Docker storage driver"""

    master_lb_enabled = wsme.wsattr(types.boolean, default=False)
    """Indicates whether created bays should have a load balancer for master
       nodes or not.
       """

    floating_ip_enabled = wsme.wsattr(types.boolean, default=True)
    """Indicates whether created bays should have a floating ip or not."""

    hidden = wsme.wsattr(types.boolean, default=False)
    """Indicates whether the Baymodel is hidden or not."""

    def __init__(self, **kwargs):
        self.fields = []
        for field in objects.ClusterTemplate.fields:
            # Skip fields we do not expose.
            if not hasattr(self, field):
                continue
            self.fields.append(field)
            setattr(self, field, kwargs.get(field, wtypes.Unset))

    @staticmethod
    def _convert_with_links(baymodel, url):
        baymodel.links = [link.Link.make_link('self', url,
                                              'baymodels', baymodel.uuid),
                          link.Link.make_link('bookmark', url,
                                              'baymodels', baymodel.uuid,
                                              bookmark=True)]
        return baymodel

    @classmethod
    def convert_with_links(cls, rpc_baymodel):
        baymodel = BayModel(**rpc_baymodel.as_dict())
        return cls._convert_with_links(baymodel, pecan.request.host_url)

    @classmethod
    def sample(cls):
        sample = cls(
            uuid='27e3153e-d5bf-4b7e-b517-fb518e17f34c',
            name='example',
            image_id='Fedora-k8s',
            flavor_id='m1.small',
            master_flavor_id='m1.small',
            dns_nameserver='8.8.1.1',
            keypair_id='keypair1',
            external_network_id='ffc44e4a-2319-4062-bce0-9ae1c38b05ba',
            fixed_network='private',
            fixed_subnet='private-subnet',
            network_driver='libnetwork',
            volume_driver='cinder',
            apiserver_port=8080,
            docker_volume_size=25,
            docker_storage_driver='devicemapper',
            cluster_distro='fedora-atomic',
            coe=fields.ClusterType.KUBERNETES,
            http_proxy='http://proxy.com:123',
            https_proxy='https://proxy.com:123',
            no_proxy='192.168.0.1,192.168.0.2,192.168.0.3',
            labels={'key1': 'val1', 'key2': 'val2'},
            server_type='vm',
            insecure_registry='10.238.100.100:5000',
            created_at=timeutils.utcnow(),
            updated_at=timeutils.utcnow(),
            public=False,
            master_lb_enabled=False,
            floating_ip_enabled=True,
            hidden=False,
        )
        return cls._convert_with_links(sample, 'http://localhost:9511')


class BayModelPatchType(types.JsonPatchType):
    _api_base = BayModel
    _extra_non_removable_attrs = {'/network_driver', '/external_network_id',
                                  '/tls_disabled', '/public', '/server_type',
                                  '/coe', '/registry_enabled',
                                  '/cluster_distro', '/hidden'}


class BayModelCollection(collection.Collection):
    """API representation of a collection of Baymodels."""

    baymodels = [BayModel]
    """A list containing Baymodel objects"""

    def __init__(self, **kwargs):
        self._type = 'baymodels'

    @staticmethod
    def convert_with_links(rpc_baymodels, limit, url=None, **kwargs):
        collection = BayModelCollection()
        collection.baymodels = [BayModel.convert_with_links(p)
                                for p in rpc_baymodels]
        collection.next = collection.get_next(limit, url=url, **kwargs)
        return collection

    @classmethod
    def sample(cls):
        sample = cls()
        sample.baymodels = [BayModel.sample()]
        return sample


class BayModelsController(base.Controller):
    """REST controller for Baymodels."""

    _custom_actions = {
        'detail': ['GET'],
    }

    def _generate_name_for_baymodel(self, context):
        """Generate a random name like: zeta-22-model."""

        name_gen = name_generator.NameGenerator()
        name = name_gen.generate()
        return name + '-model'

    def _get_baymodels_collection(self, marker, limit,
                                  sort_key, sort_dir, resource_url=None):

        limit = api_utils.validate_limit(limit)
        sort_dir = api_utils.validate_sort_dir(sort_dir)

        marker_obj = None
        if marker:
            marker_obj = objects.ClusterTemplate.get_by_uuid(
                pecan.request.context, marker)

        baymodels = objects.ClusterTemplate.list(pecan.request.context, limit,
                                                 marker_obj, sort_key=sort_key,
                                                 sort_dir=sort_dir)

        return BayModelCollection.convert_with_links(baymodels, limit,
                                                     url=resource_url,
                                                     sort_key=sort_key,
                                                     sort_dir=sort_dir)

    @expose.expose(BayModelCollection, types.uuid, int, wtypes.text,
                   wtypes.text)
    def get_all(self, marker=None, limit=None, sort_key='id',
                sort_dir='asc'):
        """Retrieve a list of Baymodels.

        :param marker: pagination marker for large data sets.
        :param limit: maximum number of resources to return in a single result.
        :param sort_key: column to sort results by. Default: id.
        :param sort_dir: direction to sort. "asc" or "desc". Default: asc.
        """
        context = pecan.request.context
        policy.enforce(context, 'baymodel:get_all',
                       action='baymodel:get_all')
        return self._get_baymodels_collection(marker, limit, sort_key,
                                              sort_dir)

    @expose.expose(BayModelCollection, types.uuid, int, wtypes.text,
                   wtypes.text)
    def detail(self, marker=None, limit=None, sort_key='id',
               sort_dir='asc'):
        """Retrieve a list of Baymodels with detail.

        :param marker: pagination marker for large data sets.
        :param limit: maximum number of resources to return in a single result.
        :param sort_key: column to sort results by. Default: id.
        :param sort_dir: direction to sort. "asc" or "desc". Default: asc.
        """
        context = pecan.request.context
        policy.enforce(context, 'baymodel:detail',
                       action='baymodel:detail')

        # NOTE(lucasagomes): /detail should only work against collections
        parent = pecan.request.path.split('/')[:-1][-1]
        if parent != "baymodels":
            raise exception.HTTPNotFound

        resource_url = '/'.join(['baymodels', 'detail'])
        return self._get_baymodels_collection(marker, limit,
                                              sort_key, sort_dir, resource_url)

    @expose.expose(BayModel, types.uuid_or_name)
    def get_one(self, baymodel_ident):
        """Retrieve information about the given Baymodel.

        :param baymodel_ident: UUID or logical name of a baymodel.
        """
        context = pecan.request.context
        baymodel = api_utils.get_resource('ClusterTemplate', baymodel_ident)
        if not baymodel.public:
            policy.enforce(context, 'baymodel:get', baymodel.as_dict(),
                           action='baymodel:get')

        return BayModel.convert_with_links(baymodel)

    @expose.expose(BayModel, body=BayModel, status_code=201)
    @validation.enforce_server_type()
    @validation.enforce_network_driver_types_create()
    @validation.enforce_volume_driver_types_create()
    @validation.enforce_volume_storage_size_create()
    def post(self, baymodel):
        """Create a new Baymodel.

        :param baymodel: a Baymodel within the request body.
        """
        context = pecan.request.context
        policy.enforce(context, 'baymodel:create',
                       action='baymodel:create')
        baymodel_dict = baymodel.as_dict()
        cli = clients.OpenStackClients(context)
        attr_validator.validate_os_resources(context, baymodel_dict)
        image_data = attr_validator.validate_image(cli,
                                                   baymodel_dict['image_id'])
        baymodel_dict['cluster_distro'] = image_data['os_distro']
        baymodel_dict['project_id'] = context.project_id
        baymodel_dict['user_id'] = context.user_id
        # check permissions for making baymodel public
        if baymodel_dict['public']:
            if not policy.enforce(context, "baymodel:publish", None,
                                  do_raise=False):
                raise exception.ClusterTemplatePublishDenied()

        # NOTE(yuywz): We will generate a random human-readable name for
        # baymodel if the name is not specified by user.
        arg_name = baymodel_dict.get('name')
        name = arg_name or self._generate_name_for_baymodel(context)
        baymodel_dict['name'] = name

        new_baymodel = objects.ClusterTemplate(context, **baymodel_dict)
        new_baymodel.create()
        # Set the HTTP Location Header
        pecan.response.location = link.build_url('baymodels',
                                                 new_baymodel.uuid)
        return BayModel.convert_with_links(new_baymodel)

    @wsme.validate(types.uuid_or_name, [BayModelPatchType])
    @expose.expose(BayModel, types.uuid_or_name, body=[BayModelPatchType])
    @validation.enforce_network_driver_types_update()
    @validation.enforce_volume_driver_types_update()
    def patch(self, baymodel_ident, patch):
        """Update an existing Baymodel.

        :param baymodel_ident: UUID or logic name of a Baymodel.
        :param patch: a json PATCH document to apply to this Baymodel.
        """
        context = pecan.request.context
        baymodel = api_utils.get_resource('ClusterTemplate', baymodel_ident)
        policy.enforce(context, 'baymodel:update', baymodel.as_dict(),
                       action='baymodel:update')
        try:
            baymodel_dict = baymodel.as_dict()
            new_baymodel = BayModel(**api_utils.apply_jsonpatch(
                baymodel_dict,
                patch))
        except api_utils.JSONPATCH_EXCEPTIONS as e:
            raise exception.PatchError(patch=patch, reason=e)

        new_baymodel_dict = new_baymodel.as_dict()
        attr_validator.validate_os_resources(context, new_baymodel_dict)
        # check permissions when updating baymodel public flag
        if baymodel.public != new_baymodel.public:
            if not policy.enforce(context, "baymodel:publish", None,
                                  do_raise=False):
                raise exception.ClusterTemplatePublishDenied()

        # Update only the fields that have changed
        for field in objects.ClusterTemplate.fields:
            try:
                patch_val = getattr(new_baymodel, field)
            except AttributeError:
                # Ignore fields that aren't exposed in the API
                continue
            if patch_val == wtypes.Unset:
                patch_val = None
            if baymodel[field] != patch_val:
                baymodel[field] = patch_val

        baymodel.save()
        return BayModel.convert_with_links(baymodel)

    @expose.expose(None, types.uuid_or_name, status_code=204)
    def delete(self, baymodel_ident):
        """Delete a Baymodel.

        :param baymodel_ident: UUID or logical name of a Baymodel.
        """
        context = pecan.request.context
        baymodel = api_utils.get_resource('ClusterTemplate', baymodel_ident)
        policy.enforce(context, 'baymodel:delete', baymodel.as_dict(),
                       action='baymodel:delete')
        baymodel.destroy()
